#!/bin/sh
# -*-Tcl-*- (for emacs) \
# if [ `id -u` -ne 0 ]; then echo "-E- Only root can run this application"; exit 1; fi; \
exec tclsh "$0" "$@"

#
# mst_ib_add :
# Parse the given lst file and print nice device names for
# mst inband access
#

# lst line format:
# { CA Ports:02 SystemGUID:0002c90001777053 NodeGUID:0002c90001777050 PortGUID:0002c90001777051 VenID:000002C9 DevID:634A0000 Rev:000000A0 {ss005 HCA-1} LID:0001 PN:01 } { SW Ports:18 SystemGUID:0000000000001230 NodeGUID:0000000000000001 PortGUID:0000000000000001 VenID:000002C9 DevID:B9240000 Rev:000001A1 {MT47396 Infiniscale-III Mellanox Technologies} LID:0003 PN:01 } PHY=4x LOG=ACT SPD=2.5

# IB_DB array
# Major key: Node guid
# Minor keys: Desc Type Dev LID

array set IB_DB {}
set gMelVendId 0x2c9

set gPrintWithGuid 0

proc isNodeIgnored {vendId} {
    global g_onlyMlnx
    global gMelVendId
    if {$g_onlyMlnx == 1 && $vendId != $gMelVendId} {
        return 1
    }
    return 0
} 

proc addNode {portData} {
    global IB_DB
    global gPrintWithGuid
    if {![ regexp { VenID:([[:xdigit:]]+)} $portData d1 vendId]} {
        error "Bad port format - Missing VenID: $portData"
    }
    set vendId [expr "0x$vendId"]
    
    if {[isNodeIgnored $vendId]} {
        return 
    }
   
    if {![ regexp { NodeGUID:([[:xdigit:]]+)} $portData d1 nguid]} {
        error "Bad port format - Missing NodeGUID: $portData"
    }
   
    if {![ regexp { PortGUID:([[:xdigit:]]+)} $portData d1 pguid]} {
        error "Bad port format - Missing PortGUID: $portData"
    }

    # If this node already exists in db - no need to re-add it
    
    if {[info exist IB_DB($nguid:NodeGUID)]} {
        set exPortGuid $IB_DB($nguid:PortGUID)
        if { $gPrintWithGuid == 0 ||
             $exPortGuid == $pguid } {
            return
        } else {
            set nguid ${nguid}_${pguid}
        }
    }

    if {![ regexp { DevID:([[:xdigit:]]+)} $portData d1 devId]} {
        error "Bad link format - Missing DevID: $portData"
    }
    # dev id seem to have 4 extra zeroes on the right.
    # HACK - ignore these zeroes.
    set devId [string range $devId 0 3]
    set devId [expr "0x$devId"]

    if {![ regexp { LID:([[:xdigit:]]+)} $portData d1 lid]} {
        error "Bad link format - Missing LID: $portData"
    }
    set lid "0x$lid"
    set lid [ format "0x%04X" $lid]
    
    set type [lindex $portData 0]
    set desc [lindex $portData 8]

    if {$type == "CA-SM"} {
            set type "CA"
    } elseif {$type == "SW-SM"} {
        set type "SW"
    }
    if {$devId != 43132} {
        set IB_DB($nguid:NodeGUID)  $nguid
        set IB_DB($nguid:PortGUID)  $pguid
        set IB_DB($nguid:LID)       $lid
        set IB_DB($nguid:Desc)      $desc
        set IB_DB($nguid:Type)      $type
        set IB_DB($nguid:DevId)     $devId
    }
}

proc isIbdrSupported {devId} {
    #                  ConnectIB              SwitchX           SwitchIB
    # return [expr $devId == 0x1011 || $devId == 0xc738 || $devId == 0xcb20]
    
    # CHECK IF THE DEVID IS NOT FOR AN OLD DEVICE THAT IS NOT SUPPORTED
    set unSupportedDevIds {0xbd34 0xbd35 0xbd36 0xfa66 0xfa7a 0x1003 0x1007
                            48436  48437  48438  64102  64122  4099   4103}
    return [expr [lsearch $unSupportedDevIds $devId] < 0]
}

proc addMstIb {hcaId port ibdr} {
    global IB_DB
    global gPrintWithGuid

    set suffix ""
    if {$hcaId != ""} {
        append suffix ",$hcaId"
        if {$port != 0} {
            append suffix ",$port"
        }
    }

    foreach nguid [array names IB_DB *:NodeGUID] {
        set ng [lindex [split $nguid ":"] 0]

        set dev "$IB_DB($ng:Type)_MT$IB_DB($ng:DevId)_"
        set desc $IB_DB($ng:Desc)
        regsub -all { } $desc "_" desc

        if {[regexp {^MF0;} $desc] } {
            # Support new managed switch node desc in format: "MF0;switch-11208a:MTS3610/S09/U1"
            # Keep switch name and Slot name in the final dev name. Loose the switch type and slot type.
            regsub      {^MF0;}  $desc ""  desc
            regsub -all {:[^/]*} $desc ""  desc
            regsub -all {/}      $desc "_" desc
            regsub      {_U1$}   $desc ""  desc
            append dev $desc
            append dev "_"
        } elseif {$IB_DB($ng:Type) == "CA"} {
            append dev $desc
            append dev "_"
        } else {
            append dev $IB_DB($ng:NodeGUID)
            append dev "_"        
        }

        if {$ibdr == 1 && [info exists IB_DB($ng:DR)] && [isIbdrSupported $IB_DB($ng:DevId)]} {
            set dr $IB_DB($ng:DR)
            regsub -all {,}      $dr "." dr
            append dev "ibdr-$dr"
        } else {
            append dev "lid-$IB_DB($ng:LID)"
        }

        # make sure we done have bad chars in the dev name
        regsub -all {[:;/]} $dev "_" dev
        
        if {$gPrintWithGuid == 1} {
            set pg $IB_DB($ng:PortGUID)
            puts "$dev$suffix#$pg"
        } else {
            puts "$dev$suffix"
        }
    }
}

proc parseLstFile {lstFile} {
    if {[catch {set fh [open $lstFile r]} e]} {
        puts stderr "-E- Failed to open file $lstFile: $e"
        exit 1
    }

    set notActiveLinksFound 0
    set lineNum 0
    while {[gets $fh sLine] >= 0} {
        incr lineNum

        if {[regexp { LOG=(\S+)} $sLine d1 portState]} {
            if {$portState != "ACT"} {
                set notActiveLinksFound 1
                continue
            }
        } else {
            continue
        }
        if {[catch {addNode [lindex $sLine 0] ; addNode [lindex $sLine 1]} e ]} {
            puts stderr "-E- $lstFile:$lineNum: Bad LST line format: $e"
            exit 1
        }
    }

    close $fh

    return $notActiveLinksFound
}

set g_nodeDiscoverIsOk 0

proc addNodeNetdiscover {nodeInfo ibdr} {
    global IB_DB
    global g_nodeDiscoverIsOk
    
    # Get vendor ID
    # vendid=0x2c9 -> 0x2c9
    if {![ regexp {vendid=(0x[[:xdigit:]]+)} $nodeInfo d1 vendId]} {
        error "Bad node format - Missing VendorId: $nodeInfo"
    }
    
    if {[isNodeIgnored $vendId]} {
        return 
    }
 
    # Get node guid
    # switchguid=0x2c902004177d8(2c902004177d8) -> 0x2c902004177d8
    if {![ regexp {(switchguid|caguid)=(0x[[:xdigit:]]+)} $nodeInfo d1 d2 nguid]} {
        error "Bad node format - Missing NodeGuid: $nodeInfo"
    }   
    
    # If this node already exist in db - no need to re-add it
    if {[info exist IB_DB($nguid:NodeGUID)]} {
        return
    }
    # Get the device ID
    # devid=0xbd36 -> 0xbd36
    if {![ regexp {devid=(0x[[:xdigit:]]+)} $nodeInfo d1 devId]} {
        error "Bad node format - Missing Device ID: $nodeInfo"
    }
    set devId [ format "%d" $devId]
    # Ignore InfiniScale device ID.
    if {$devId == 43132} {
        return 
    }
    
    # Get the type
    # Switch  36 "S-0002c902004177d8"         # "MF0;appssw02:IS5035/U1" -> Switch
    # Ca      2 "H-2c90300300111bb0"          # "MF0;appsbx01:BX9020/U1" -> Ca
    if {![ regexp  {(Ca|Switch)(.)*# (\".*\")(.*)} $nodeInfo typemsg type]} {
        error "Bad node format - Missing type: $nodeInfo"
    }  
    
    # Get the description 
    # Switch  36 "S-0002c902004177d8"         # "MF0;appssw02:IS5035/U1" -> # "MF0;appssw02:IS5035/U1"
    set desc1  [string range $typemsg [string first \# $typemsg 0] end]
    
    set index1 [expr [string first \" $desc1 1] + 1]
    set index2 [expr [string first \" $desc1 $index1] - 1]
    # # "MF0;appssw02:IS5035/U1" -> MF0;appssw02:IS5035/U1
    
    set desc   [string range $desc1 $index1 $index2]
    
    if { $type == "Switch" } {
        set type "SW"        
        # Get lid for Switch 
        # Switch  36 "S-0002c90200415190"         # "MF0;appssw01:IS5035/U1" enhanced port 0 lid 11 lmc 0
        if {![ regexp  {Switch.*port 0 lid ([0-9]+)} $nodeInfo d1 lid]} {
            error "Bad node format - Missing lid: $nodeInfo"
        }
        set pguid $nguid
    } else {
        set type "CA"  
        # Get lid for HCA
        # [1](2c90300300111bb1)   "S-0002c902004177d8"[6]         # lid 12 lmc 0 "MF0;appssw02:IS5035/U1" lid 13 4xQDR ->      # lid 12 lmc 0 "MF0;appssw02:IS5035/U1" lid 13 4xQDR
        if {![ regexp  {\[[1-2]\]\(([[:xdigit:]]+)\) * \"(H|S)\-[[:xdigit:]]+\"\[[0-9]+\](.*) lmc}  $nodeInfo d1 pguid d2 lidStr]} {
                       
            if { ${g_nodeDiscoverIsOk} == 0 } {
                error "Failed to get first lid: Maybe the port you scan is down"
            } else {
                error "Bad node format - Missing lid info : $nodeInfo"
            }
            
        }
        set lidStr [string trim $lidStr]
        #      # lid 12 lmc 0 "MF0;appssw02:IS5035/U1" lid 13 4xQDR -> 12
        if {![ regexp  {\# lid ([0-9]+)} $lidStr d1 lid]} {
            error "Bad node format - Missing lid: $nodeInfo"
        }
    }  
    set lid [ format "0x%04X" $lid]
    set legalDevId 0
    if {$ibdr == 1 &&  [isIbdrSupported $devId]} {
        set legalDevId 1
    } elseif {$lid != 0 && $lid != 0xFFFF} {
        set legalDevId 1
    }
    if {$legalDevId == 1} {
        set IB_DB($nguid:NodeGUID)  $nguid
        set IB_DB($nguid:PortGUID)  $pguid
        set IB_DB($nguid:LID)       $lid
        set IB_DB($nguid:Desc)      $desc
        set IB_DB($nguid:Type)      $type
        set IB_DB($nguid:DevId)     $devId
        set g_nodeDiscoverIsOk 1
    }
}

proc filter_nonprintable_chars {string} {
    set length [string length $string]
    set new_string ""
    for {set index 0} {$index < $length} {incr index} {
        set char [string index $string $index]
        if {[string is print $char]} {
            set new_string "$new_string$char"
        }
    }
    return $new_string
}

proc parseNetdiscoverFile {outFile ibdr} {
    global IB_DB

    if {[catch {set fh [open $outFile r]} e]} {
        puts stderr "-E- Failed to open file $outFile: $e"
        exit 1
    }
    set nodeInfo ""
    set lineNum 0
    while {[gets $fh sLine] >= 0} {
        incr lineNum

        #DR path slid 0; dlid 0; 0 -> new Channel Adapter {0000000000001234} portnum 1 lid 0-0 ""
        #DR path slid 0; dlid 0; 0,1 -> new Switch {0002c903004693c1} portnum 0 lid 0-0 ""
        if {[regexp {^DR path slid 0; dlid 0; ([0-9,]+) -> new .+ \{([[:xdigit:]]+)\}} $sLine d1 drPath ng]} {
            if {![regexp {^0x} $ng]} {
                set ng "0x[string trimleft $ng 0]"
            }
            set IBDR_DB($ng) $drPath
        }

        #puts "-D- vendid = $sLine"
        if {[regexp {vendid=} $sLine]} {
            set nodeInfo "$sLine"
            while {[gets $fh sLine] >= 0} {
                
                if { $sLine == "" } {
                    if {[catch {addNodeNetdiscover $nodeInfo $ibdr} e ]} {
                        puts stderr "-E- $outFile:$lineNum: Bad ibnetdiscover output format: $e"
                        exit 1
                    }
                    set nodeInfo ""
                    break
                }
                # Collect the nodeinfo line by line into one line.
                set newLine [filter_nonprintable_chars $sLine]
                set nodeInfo "$nodeInfo $newLine"
            }
            # We got here when we have the last nodeInfo in the topo file.
            if {$nodeInfo != ""} {
                 if {[catch {addNodeNetdiscover $nodeInfo $ibdr} e ]} {
                    puts stderr "-E- $outFile:$lineNum: Bad ibnetdiscover output format (Last node): $e"
                    exit 1
                }
            }
            
        }
    }

    # merge DR data indo IB_DB
    foreach nguid [array names IB_DB *:NodeGUID] {
        set ng [lindex [split $nguid ":"] 0]

        if {[info exist IBDR_DB($ng)]} {
            set IB_DB($ng:DR) $IBDR_DB($ng)
        }
    }

    close $fh
    return 0  
}    
    


proc main {argv0 argv arg} {
    global g_onlyMlnx
    global gPrintWithGuid
    set g_onlyMlnx 1
    set usage "  Usage: $argv0 <lst_file> <only_mlnx|all> \[file_type\] \[ibdr 0|1\] \[hca_id\] \[exit_port_num\]"
    

    if {[lindex $argv 0] == "-h" || [lindex $argv 0] == "--help"} {
        puts stdout $usage
        exit 0	
	} elseif {[llength $argv] > 7 || [llength $argv] < 2} {
        puts stderr $usage
        exit 1
	}

    set hcaId ""
    set portNum 0
    set diagType "diagnet"
    set fileType $diagType
    
    set lstFile  [lindex $argv 0]
    set nodeType [lindex $argv 1]
    set ibdr 0
    
    if {$nodeType == "all"} {
        set g_onlyMlnx 0
    }

    # Get user flags
    if {[llength $argv] > 2} {
        set fileType [lindex $argv 2]
    }

    if {[llength $argv] > 3} {
        set ibdr [lindex $argv 3]
    }   

    if {[llength $argv] > 4} {
        set hcaId [lindex $argv 4]
    }
    if {[llength $argv] > 5} {
        set portNum [lindex $argv 5]
    }
    if {[llength $argv] > 6} {
        if {[lindex $argv 6] == "--with-guids"} {
            set gPrintWithGuid 1
        } else {
            puts stderr "  Usage: $argv0 <lst_file> <only_mlnx|all> \[file_type\] \[ibdr 0|1\] \[hca_id\] \[exit_port_num\] \[--with-guids\]"
            exit 1
        }
    }
    # Parse files
    if {$fileType == $diagType} {
        set rc [parseLstFile $lstFile]
    } else {
        set rc [parseNetdiscoverFile $lstFile $ibdr]
    }

    addMstIb $hcaId $portNum $ibdr

    if {$rc != 0} {
        puts stderr "-W- Non active links found in the IB fabric. These links are not added to the device list. Make sure that a subnet-manager is running."
    }
}

main $argv0 $argv $argc
